/*
 * Copyright 2012 metArch
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.metarch.timer.model;

import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.Map;
import java.util.Set;
import java.util.TreeSet;

import org.metarch.timer.model.Stopwatch.State;
import org.metarch.timer.reports.TimerConsoleReport;
import org.metarch.timer.util.TimerComparator;

/**
 * The CodeTimer provides a set of static method that will simplify performance
 * testing your code.
 * 
 * Be warned: using static methods creates a strong dependency with this library
 * so avoid using it in production code libraries. Use it either in your testing
 * code or remove it afterwards.
 * 
 * @author Niels Krijger <niels@kryger.nl>
 */
public class CodeTimer {

	private static boolean enabled = true;
	private static Map<String, Timer> timers = new LinkedHashMap<>();

	/**
	 * Currently running timers ordered from longest running to shortest
	 * running.
	 */
	private static LinkedList<Timer> activeTimers = new LinkedList<>();

	/**
	 * Creates a new timer with the calling method as tag name. Be aware, this
	 * method does not retrieve the calling method's parameters due to the
	 * limited information unavailable in the stacktrace used to retrieve
	 * the calling method. To avoid problems with overloaded methods the
	 * linenumber is appended to the tag.
	 * 
	 * @return the generated tag name or null when code timing has been
	 *         disabled.
	 */
	public synchronized static String startTimer() {
		if (isCodeTimingEnabled()) {
			Throwable throwable = new Throwable();
			StackTraceElement stackTrace = throwable.getStackTrace()[1];
			String lineNumber = "" + stackTrace.getLineNumber();
			String fullClassName = stackTrace.getClassName().toString();
			String method = stackTrace.getMethodName();
			String tag = String.format("%1$s.%2$s (line #%3$s)", fullClassName, method, lineNumber);
			startTimer(tag);
			return tag;
		} else {
			return null;
		}
	}

	/**
	 * Starts a new timer with given id. If the timer tag already exists the
	 * previous timer with that tag is retrieved, any running stopwatches are
	 * stopped and a new stopwatch is started.
	 * 
	 * If you want overlapping timers (say, one for the entire code block and
	 * one for a smaller inner piece), use timers with different tags.
	 * 
	 * @param tag
	 *            the timer's tag.
	 */
	public synchronized static void startTimer(String tag) {
		startTimer(tag, null);
	}

	/**
	 * Starts a new timer with a given tag. If the timer already exists the
	 * instance is retrieved, any running stopwatches are stopped and a new
	 * stopwatch is started.
	 * 
	 * When code timing is disabled nothing happens.
	 * 
	 * @param tag
	 *            the timer's tag Should be unique.
	 * @param description
	 *            the timer's description
	 */
	public synchronized static void startTimer(String tag, String description) {
		if (isCodeTimingEnabled()) {
			Timer timer = null;
			if (timers.containsKey(tag)) {
				timer = timers.get(tag);
				if (activeTimers.contains(timer)) {
					activeTimers.remove(timer);
				}
			} else {
				timer = new Timer(tag, description);
				timers.put(tag, timer);
			}
			activeTimers.add(timer);
			getLastActiveTimer().start();
		}
	}

	/**
	 * Stops the timer that was last started.
	 * 
	 * When code timing is disabled nothing happens.
	 * 
	 * @throws IllegalArgumentException
	 *             thrown when no timer was started, paused or resumed prior to
	 *             calling stopTimer().
	 */
	public synchronized static void stopTimer() {
		if (isCodeTimingEnabled()) {
			if (getLastActiveTimer() != null) {
				stopTimer(getLastActiveTimer().getTag());
			} else {
				throw new IllegalStateException("No known timer to stop, "
						+ "make sure at least one timer has been started");
			}
		}
	}

	/**
	 * Stops a running timer.
	 * 
	 * When code timing is disabled nothing happens.
	 * 
	 * @param tag
	 *            the unique identifier of the timer.
	 * @throws IllegalStateException
	 *             when the timer requested to stop isn't running.
	 * @throws IllegalArgumentException
	 *             when no timer with given tag was found.
	 */
	public synchronized static void stopTimer(String tag) {
		if (isCodeTimingEnabled()) {
			if (timers.containsKey(tag)) {
				Timer timer = timers.get(tag);
				timer.stop();
				activeTimers.remove(timer);
			} else {
				throw new IllegalArgumentException("Timer with tag '" + tag
						+ "' " + "could not be found");
			}
		}
	}

	/**
	 * Pauses the timer that was last started or resumed.
	 * 
	 * When code timing is disabled nothing happens.
	 * 
	 * @throws IllegalArgumentException
	 *             thrown when no timer was paused or resumed prior to calling
	 *             stopTimer().
	 */
	public synchronized static void pauseTimer() {
		if (isCodeTimingEnabled()) {
			if (!getRunningTimers().isEmpty()) {
				pauseTimer(getLastActiveTimer().getTag());
			} else {
				throw new IllegalStateException("No known timer to pause, "
						+ "make sure a timer is running");
			}
		}
	}

	/**
	 * Pauses a running timer.
	 * 
	 * When code timing is disabled nothing happens.
	 * 
	 * @param tag
	 *            the unique identifier of the timer.
	 * @throws IllegalStateException
	 *             when the timer requested to stop isn't running.
	 * @throws IllegalArgumentException
	 *             when no timer with given tag was found.
	 */
	public synchronized static void pauseTimer(String tag) {
		if (isCodeTimingEnabled()) {
			if (timers.containsKey(tag)) {
				Timer timer = timers.get(tag);
				timer.pause();
				activeTimers.remove(timer);
				activeTimers.add(timer);
			} else {
				throw new IllegalArgumentException("Timer with tag '" + tag
						+ "' " + "could not be found");
			}
		}
	}

	/**
	 * Resumes the timer that was last started, resumed or paused. Mind you, if
	 * the timer is not paused throws an exception.
	 * 
	 * When code timing is disabled nothing happens.
	 * 
	 * @throws IllegalStateException
	 *             thrown when no timer was paused prior to calling
	 *             resumeTimer().
	 */
	public synchronized static void resumeTimer() {
		if (isCodeTimingEnabled()) {
			if (getLastActiveTimer() != null) {
				resumeTimer(getLastActiveTimer().getTag());
			} else {
				throw new IllegalStateException("No known timer to pause, "
						+ "make sure a timer is running");
			}
		}
	}

	/**
	 * Pauses a running timer.
	 * 
	 * When code timing is disabled nothing happens.
	 * 
	 * @param tag
	 *            the unique identifier of the timer.
	 * @throws IllegalStateException
	 *             when the timer requested to stop isn't running.
	 * @throws IllegalArgumentException
	 *             when no timer with given tag was found.
	 */
	public synchronized static void resumeTimer(String tag) {
		if (isCodeTimingEnabled()) {
			if (timers.containsKey(tag)) {
				Timer timer = timers.get(tag);
				timer.resume();
			} else {
				throw new IllegalArgumentException("Timer with tag '" + tag
						+ "' " + "could not be found");
			}
		}
	}

	/**
	 * Returns the timer to which the specified tag is mapped, or null if no
	 * timer with specified tag could be found.
	 * 
	 * @param tag
	 * @return Timer with specified tag, or null if no timer could be found or
	 *         code timing is disabled.
	 */
	public synchronized static Timer getTimer(String tag) {
		return (isCodeTimingEnabled()) ? timers.get(tag) : null;
	}

	/**
	 * Returns the last running or paused timer.
	 * 
	 * @return The last running or paused timer, or null if no running or paused
	 *         timer exist.
	 */
	public synchronized static Timer getLastActiveTimer() {
		return activeTimers.get(activeTimers.size() - 1);
	}

	/**
	 * Returns all known timers.
	 * 
	 * @return set with timers in the order they were created.
	 */
	public synchronized static Collection<Timer> getAllTimers() {
		if (isCodeTimingEnabled()) {
			return timers.values();
		} else {
			return Collections.emptySet();
		}
	}

	/**
	 * Returns all known timers sorted by a comparator.
	 * 
	 * Several comparator are already available in
	 * {@link meta.timer.util.TimerComparator}.
	 * 
	 * @return sorted set of timers or an empty set if code timing is disabled.
	 */
	public synchronized static Collection<Timer> getAllTimers(Comparator<Timer> comparator) {
		if (isCodeTimingEnabled()) {
			Set<Timer> result = new TreeSet<>(comparator);
			result.addAll(timers.values());
			return result;
		} else {
			return Collections.emptySet();
		}
	}

	/**
	 * Stops all running timers. This method loops through all known timers,
	 * checks whether they're running and if so stops them.
	 * 
	 * @ensure <code>getRunningTimers().size() == 0</code>
	 */
	public synchronized static void stopAllTimers() {
		if (isCodeTimingEnabled()) {
			Iterator<Timer> iterator = timers.values().iterator();
			while (iterator.hasNext()) {
				Timer timer = iterator.next();
				if (timer.isRunning()) {
					timer.stop();
				} else if (timer.isPaused()) {
					timer.stop();
				}
			}
		}
	}

	/**
	 * Pauses all running timers. This method loops through all known timers,
	 * checks whether they're running and if so pauses them.
	 * 
	 * @ensure <code>getRunningTimers().size() == 0</code>
	 */
	public synchronized static void pauseAllTimers() {
		if (isCodeTimingEnabled()) {
			Iterator<Timer> iterator = timers.values().iterator();
			while (iterator.hasNext()) {
				Timer timer = iterator.next();
				if (timer.isRunning()) {
					timer.pause();
				}
			}
		}
	}

	/**
	 * Enables code timing.
	 * 
	 * Enabling or disabling does not affect the inner state of the CodeTimer.
	 */
	public synchronized static void enableCodeTiming() {
		enabled = true;
	}

	/**
	 * Disables code timing.
	 * 
	 * Enabling or disabling does not affect the inner state of the CodeTimer.
	 * Results tracked prior to disabling the code timer will be kept. If you
	 * want to erase all previous results, use {@link #reset()}.
	 * 
	 * Because disabling code timing does not affect any inner state but merely
	 * disables changing state afterwards, timers that were running prior to
	 * disabling code timing will remain running until code timing is re-enabled
	 * and the timers are stopped. Often this will be undesirable.
	 * 
	 */
	public synchronized static void disableCodeTiming() {
		enabled = false;
	}

	/**
	 * Returns whether the CodeTimer is enabled or not. When the CodeTimer is
	 * disabled no times are being tracked.
	 * 
	 * @return true if the code timer keeps track of times, otherwise false.
	 */
	public synchronized static boolean isCodeTimingEnabled() {
		return enabled;
	}

	/**
	 * A list with all timers currently running. If no timers are currently
	 * running returns an empty set.
	 * 
	 * @return set with running timers.
	 */
	public synchronized static Collection<Timer> getRunningTimers() {
		return getTimers(State.RUNNING);
	}

	/**
	 * A list with all timers currently paused . If no timers are currently
	 * paused returns an empty set.
	 * 
	 * @return set with paused timers.
	 */
	public synchronized static Collection<Timer> getPausedTimers() {
		return getTimers(State.PAUSED);
	}

	/**
	 * Returns list with all timers in specified mode.
	 * 
	 * If no timers are currently in the specified mode returns an empty list.
	 * 
	 * @param mode
	 * @return
	 */
	private synchronized static Collection<Timer> getTimers(State mode) {
		if (isCodeTimingEnabled()) {
			Set<Timer> result = new LinkedHashSet<>();
			for (Timer timer : timers.values()) {
				switch (mode) {
				case PAUSED:
					if (timer.isPaused()) {
						result.add(timer);
					}
					break;
				case RUNNING:
					if (timer.isRunning()) {
						result.add(timer);
					}
					break;
				case STOPPED:
				case NOT_STARTED:
				default:
					assert false;
					break;
				}
			}
			return result;
		} else {
			return Collections.emptySet();
		}
	}

	/**
	 * Resets all timers and removes them from the codetimer.
	 * 
	 * Reset will also work while code timing is disabled. If reset is called
	 * while code timing is disabled the code timing remains disabled
	 * afterwards.
	 */
	public synchronized static void reset() {
		for (Timer timer : timers.values()) {
			timer.reset();
		}
		timers = new HashMap<>();
	}
	
	/**
	 * Generates a simple report with all known timers. The timers are sorted 
	 * by total time in descending order.
	 * 
	 * This method uses the {@link TimerConsoleReport} to format the output. 
	 * 
	 * This method is meant to be used merely in simple testing situations.
	 * For more advanced reporting you're advised to write your own
	 * report generator.
	 * 
	 * @return Formatted list of all known timers.
	 */
	public static String generateReport() {
		TimerConsoleReport report = new TimerConsoleReport();
        
        // A report is populated with a collection of all timers
        return report.generateReport(
        		getAllTimers(TimerComparator.sortTotalTimeDescending()));
	}
}