/*
 * Copyright 2012 metArch
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.metarch.timer.model;

/**
 * A stopwatch simply keeps track of how much time has passed in nanoseconds. 
 * 
 * Creating and managing stopwatches yourself is not recommended, use the Timer 
 * and CodeTimer classes instead which provide additional features.
 * 
 * @author Niels Krijger <niels@kryger.nl>
 */
public class Stopwatch {
    
    /**
     * State contains all possible states the stopwatch can be in.
     */
    public static enum State {
        NOT_STARTED,
        RUNNING,
        PAUSED,
        STOPPED
    }
    
    private long startTime;
    private long stopTime;
    private long duration;
    private State state;
    
    /**
     * Creates a new stopwatch.
     */
    public Stopwatch() {
        this.state = State.NOT_STARTED;
    }

    /**
     * Start the stopwatch.
     * 
     * @throws IllegalStateException thrown when start() is invoked while the 
     *      stopwatch <code>isNotStarted() == false</code>.
     */
    public void start() {
        if (is(State.NOT_STARTED)) {
            this.state = State.RUNNING;
            this.startTime = System.nanoTime();
        } else {
            throw new IllegalStateException("Prior to starting the stopwatch its "
                    + "state should be " + State.NOT_STARTED +" (currently "
                    + this.state + ").");
        }
    }
    
    /**
     * Stop the stopwatch while it is running or paused.
     * 
     * @throws IllegalStateException thrown when stop() is invoked while the 
     *      stopwatch <code>isRunning() == false && isPaused() == false</code>.
     */
    public void stop() {
        if (is(State.RUNNING) || is(State.PAUSED)) {
            this.stopTime = System.nanoTime();
            this.state = State.STOPPED;
            this.duration += this.stopTime - this.startTime;
        } else {
            throw new IllegalStateException("Prior to starting the stopwatch its "
                    + "state should be " + State.NOT_STARTED +" (currently "
                    + this.state + ").");
        }
    }
    
    /**
     * Pauses the stopwatch if the stopwatch is currently running. 
     * 
     * @throws IllegalStateException thrown when pause() is invoked while the 
     *      stopwatch <code>isRunning() == false</code>.
     */
    public void pause() {
        if (is(State.RUNNING)) {
            this.duration += System.nanoTime() - this.startTime;
            this.state = State.PAUSED;
        } else {
            throw new IllegalStateException("Prior to pausing the stopwatch its "
                    + "state should be " + State.RUNNING +" (currently "
                    + this.state + ").");
        }
    }
    
    /**
     * Resumes the stopwatch.
     * 
     * @throws IllegalStateException thrown when resume() is invoked while the 
     *      stopwatch <code>isPaused() == false</code>.
     */
    public void resume() {
        if (is(State.PAUSED)) {
            this.state = State.RUNNING;
            this.startTime = System.nanoTime();
        } else {
            throw new IllegalStateException("Prior to resuming the stopwatch its "
                    + "state should be " + State.PAUSED +" (currently "
                    + this.state + ").");
        }
    }
    
    /**
     * Returns in which state the stopwatch is. 
     * 
     * @return returns the {@link Stopwatch.State}.
     */
    public State currentState() {
        return this.state;
    }
    
    /**
     * Checks whether the stopwatch is in given state.
     * 
     * @param state
     * @return true if stopwatch is in given state, otherwise false.
     */
    public boolean is(State state) {
        return this.state == state;
    }
    
    /**
     * Returns the stopwatch state.
     * 
     * @return One of {@link Stopwatch.State} values.
     */
    public State getState() {
        return this.state;
    }
    
    /**
     * Returns whether stopwatch has been started or not.
     * 
     * Does the same as:
     * <pre>
     * myStopwatch.is(Stopwatch.State.NOT_STARTED)
     * </pre>
     * 
     * @return true if stopwatch has not been started yet, otherwise false.
     */
    public boolean isNotStarted() {
        return is(State.NOT_STARTED);
    }
    
    /**
     * Returns whether stopwatch is running or not. 
     * 
     * Does the same as:
     * <pre>
     * myStopwatch.is(Stopwatch.State.RUNNING)
     * </pre>
     * 
     * @return true if stopwatch is currently running, otherwise false.
     */
    public boolean isRunning() {
        return is(State.RUNNING);
    }
    
    /**
     * Returns whether stopwatch is paused or not. 
     * 
     * Does the same as:
     * <pre>
     * myStopwatch.is(Stopwatch.State.PAUSED)
     * </pre>
     * 
     * @return true if stopwatch is currently paused, otherwise false.
     */
    public boolean isPaused() {
        return is(State.PAUSED);
    }
    
    /**
     * Returns whether stopwatch has been stopped or not. 
     * 
     * Does the same as:
     * <pre>
     * myStopwatch.is(Stopwatch.State.STOPPED)
     * </pre>
     * 
     * @return true if stopwatch is currently stopped, otherwise false.
     */
    public boolean isStopped() {
        return is(State.STOPPED);
    }
    
    /**
     * Returns time the stopwatch has been running in nanoseconds.
     * 
     * @return number of nanoseconds the stopwatch has been running.
     */
    public long getTime() {
        if (isRunning()) {
            return this.duration + System.nanoTime() - this.startTime;
        } else {
            return this.duration;
        }
    }

    @Override
    public String toString() {
        return "Stopwatch{" + "startTime=" + startTime + ", stopTime=" + stopTime + ", state=" + state + '}';
    }
}
