/*
 * Copyright 2012 metArch
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.metarch.timer.model;

import java.util.LinkedList;
import java.util.List;

import org.metarch.timer.model.Stopwatch.State;

/**
 * A timer is useful when you want to start and stop several intervals. It
 * tracks this by creating and managing several stopwatches.
 * 
 * Even if you're not planning to track more than one interval the Timer is still
 * useful for its tag, description and output formats allowing you to create
 * detailed logs.
 * 
 * An example use:
 * <pre>
 * Timer timer = new Timer("My Timer", "This just keeps track of time");
 * timer.start();
 * timer.stop();
 * System.out.println("Total time: " + timer.getTotalTime() + " with an average "
 *      + " of " + timer.getAverageTime() + ".");
 * </pre>
 * 
 * This class implements the {@link Comparable} interface which compares
 * total running time, see {@link #compareTo(java.lang.Object)}.
 * 
 * @author Niels Krijger <niels@kryger.nl>
 */
public class Timer implements Comparable<Timer> {
        
    private String tag;
    private String description;
    private Stopwatch active;
    private List<Stopwatch> stopwatches = new LinkedList<>();

    /**
     * Create a new timer with a tag. It's description will remain empty.
     * 
     * @param tag 
     */
    public Timer(String tag) {
        this.tag = tag;
    }
    
    /**
     * Creates a new timer with a tag and description.
     * 
     * @param tag
     * @param description 
     */
    public Timer(String tag, String description) {
        this(tag);
        this.description = description;
    }
    
    /**
     * Returns the tag of this Timer. While not enforced by this class it is
     * highly recommended to give each Timer a unique tag to easily recognize
     * and search them in your codebase.
     * 
     * @return the timer's tag.
     */
    public String getTag() {
        return this.tag;
    }
    
    /**
     * Returns the state of the current active stopwatch.
     * 
     * @return returns the state of the active stopwatch or null if no 
     *      active stopwatch exists.
     */
    public State getState() {
        return (this.active != null) ? this.active.getState() : null;
    }
    
    /**
     * Returns the description of this Timer.
     * 
     * @return the timer's description.
     */
    public String getDescription() {
        return this.description;
    }

    /**
     * Stops any running timer and starts a new one.
     * 
     * @return 
     */
    public void start() {
        this.stop();
        Stopwatch stopwatch = new Stopwatch();
        stopwatches.add(stopwatch);
        this.active = stopwatch;
        this.active.start();
    }
    
    /**
     * Pauses currently active timer. 
     * 
     * @throws IllegalStateException when timer is not running.
     * @ensure <code>this.isPaused() == true</code>
     */
    public void pause() {
        this.active.pause();
    }
    
    /**
     * Resumes a paused timer.
     * 
     * @throws IllegalStateException when timer is not paused.
     */
    public void resume() {
        this.active.resume();
    }
    
    /**
     * Returns a list of all stopwatches managed by this timer. The first
     * stopwatch in the list is the oldest, the last stopwatch the latest.
     * 
     * This method does not start or stop any stopwatch, meaning the last
     * stopwatch might still be running when you invoke this method. You'll
     * have to stop the stopwatch manually using {@link #stop()}. 
     * 
     * @return
     */
    public List<Stopwatch> getAll() {
        return stopwatches;
    }
    
    /**
     * The number of times {@link #start()} has been called.
     * 
     * @return the number of times {@link #start()} has been called. Be aware
     *      the last stopwatch may still be running.
     */
    public int getTimesCalled() {
        return stopwatches.size();
    }
    
    /**
     * Stops any running stopwatch. Returns false if no running stopwatch could
     * be found.
     * 
     * @return false if no running stopwatch could be found.
     * 
     * @throws IllegalStateException thrown when stop() is invoked while the 
     *      timer is neither running nor paused.
     */
    public boolean stop() {
        if (this.active != null) {
            this.active.stop();
            // Remove quick reference, it remains available in the stopwatches list
            this.active = null; 
            return true;
        } else {
            return false;
        }
    }
    
    /**
     * Returns the last time tracked by this Timer in nanoseconds.
     * 
     * @return the last time tracked by this Timer or 0 if the timer has not
     *      started yet.
     */
    public long getLastTime() {
        long result = 0;
        if (this.stopwatches.size() > 0) {
            result = this.stopwatches.get(this.stopwatches.size() - 1).getTime();
        } 
        return result;
    }
    
    /**
     * Returns the longest time period in nanoseconds this Timer has been running.
     * 
     * @return the longest time period in nanoseconds this Timer has been running
     *      or 0 if the Timer has never been started.
     */
    public long getMaxTime() {
        long result = 0;
        for (Stopwatch stopwatch: stopwatches) {
            if (stopwatch.getTime() > result) {
                result = stopwatch.getTime();
            }
        }
        return result;
    }
    
    /**
     * Returns the shortest time period in nanoseconds this Timer has been running.
     * 
     * @return the shortest time period in nanoseconds this Timer has been running
     *      or 0 if the Timer has never been started.
     */
    public long getMinTime() {
        long result = 0;
        for (Stopwatch stopwatch: stopwatches) {
            if (result == 0 || stopwatch.getTime() < result) {
                result = stopwatch.getTime();
            }
        }
        return result;
    }
    
    /**
     * Returns the entire time period in nanoseconds this Timer has been running.
     * 
     * @return the entire time period in nanoseconds this Timer has been running
     *      or 0 if the Timer has never been started.
     */
    public long getTotalTime() {
        long result = 0;
        for (Stopwatch stopwatch: stopwatches) {
            result += stopwatch.getTime();
        }
        return result;
    }
    
    /**
     * Returns the average time period in nanoseconds.
     * 
     * @param precision number of digits after the decimal point.
     * @return 
     */
    public double getAverageTime() {
        return Math.round((double)getTotalTime() / (double)getTimesCalled());
    }
    
    /**
     * Removes all tracked times and stops any running stopwatches. This
     * effectively clears the Timer's state.
     * 
     * @ensure isRunning() == false
     * @ensure getAll().equals(Collections.EMPTY_LIST)
     */
    public void reset() {
        this.active = null;
        this.stopwatches = new LinkedList<>();
    }

    /**
     * Returns whether a stopwatch is currently running.
     * 
     * @return True if a stopwatch is currently running, otherwise false.
     */
    public boolean isRunning() {
        return (active != null) ? active.isRunning() : false;
    }
    
    /**
     * Returns whether current active stopwatch is currently paused.
     * 
     * @return True if a stopwatch is currently paused, otherwise false.
     */
    public boolean isPaused() {
        return (active != null) ? active.isPaused() : false;
    }

    @Override
    public String toString() {
        return "Timer{" + "tag=" + tag + ", description=" + description + ", active=" + active + ", stopwatches=" + stopwatches + '}';
    }

    /**
     * Compares two timers by their tag in alphabetic order.
     * 
     * @param timer
     * @return 
     */
    @Override
    public int compareTo(Timer timer) {
        return getTag().compareTo(timer.getTag());
    }
    
    @Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + ((tag == null) ? 0 : tag.hashCode());
		return result;
	}
    
    /**
     * Two Timer objects will be considered equal if
     * <code>this.getTag().equals(obj.getTag())</code>
     * 
     * @param obj
     */
    @Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		Timer other = (Timer) obj;
		if (tag == null) {
			if (other.tag != null)
				return false;
		} else if (!tag.equals(other.tag))
			return false;
		return true;
	}
}
