/*
 * Copyright 2012 metArch
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.metarch.timer.reports;

import java.util.Collection;
import java.util.Formatter;

import org.metarch.timer.model.Timer;
import org.metarch.timer.util.TimerFormatter;
import org.metarch.timer.util.TimerReport;

/**
 * The TimerConsoleReport formats the timers as a human readable plain text 
 * useful as console output.
 * 
 * @author Niels Krijger <niels@kryger.nl>
 */
public class TimerConsoleReport implements TimerReport {

    public static final int PRECISION = 6;
    private TimerFormatter timerFormatter;

    public TimerConsoleReport() {
        this.timerFormatter = new TimerConsoleFormatter(PRECISION);
    }

    public String getTableHeader() {
        StringBuilder output = new StringBuilder();
        Formatter formatter = new Formatter(output);
        formatter.format("%-40s %-40s %12s %7s %12s %12s %12s %-15s",
                "Tag", "Description", "Total", "Called", "Avg",
                "Max", "Min", "Current state");
        formatter.close();
        output.append("\n")
                .append(repeat("-", 40))
                .append(" ")
                .append(repeat("-", 40))
                .append(" ")
                .append(repeat("-", 12))
                .append(" ")
                .append(repeat("-", 7))
                .append(" ")
                .append(repeat("-", 12))
                .append(" ")
                .append(repeat("-", 12))
                .append(" ")
                .append(repeat("-", 12))
                .append(" ")
                .append(repeat("-", 15));
        return output.toString();
    }

    @Override
    public String generateReport(Collection<Timer> timers) {
        StringBuilder sb = new StringBuilder();
        sb.append(getTableHeader());
        sb.append("\n");
        for (Timer timer : timers) {
            sb.append(timerFormatter.format(timer)).append("\n");
        }
        return sb.toString();
    }
    
    private static String repeat(String string, int repeat) {
        String output = "";
        for (int i = 0; i < repeat; i++) {
            output += string;
        }
        return output;
    }

    private class TimerConsoleFormatter implements TimerFormatter {

        public static final String ELLIPSIS = "...";
        
		private int precision;

        public TimerConsoleFormatter(int precision) {
            this.precision = precision;
        }

        @Override
        public String format(Timer timer) {
            StringBuilder output = new StringBuilder();
            Formatter formatter = new Formatter(output);
            formatter.format("%-40s %-40s %12." + this.precision +"f " +
            		"%7d %12." + this.precision +"f %12." + this.precision + "f " +
            		"%12." + this.precision + "f %-15s",
                    ellipsis(timer.getTag(), 40),
                    (timer.getDescription() == null) ? "-" : ellipsis(timer.getDescription(), 40),
                    (float) ((float) timer.getTotalTime() / (float) 1000000000),
                    timer.getTimesCalled(),
                    (float) ((float) timer.getAverageTime() / (float) 1000000000),
                    (float) ((float) timer.getMaxTime() / (float) 1000000000),
                    (float) ((float) timer.getMinTime() / (float) 1000000000),
                    (timer.getState() == null) ? "None" : timer.getState().toString());
            formatter.close();
            return output.toString();
        }
        
        private String ellipsis(String text, int maxLength) {
            if (text.length() > maxLength) {
                text = text.substring(0, maxLength-ELLIPSIS.length());
                text += ELLIPSIS;
            }
            return text;
        }
    }
}
